#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Command line utilities dpkg and apt
#

import os
import sys
import utils
import logging
from subprocess import Popen, PIPE, STDOUT
from generalexceptions import ExternalProcessError
from string import Template

MAX_OUTPUT_LOGGING = 60 * 20

class CmdLineUtil(object):

    def __init__(self, binary):
        self.binary = binary
        self.retcodeMap = {}

    def __call__(self, cmdstr, quiet=False):
        cmdstr = "%s %s" % (self.binary, cmdstr)
        logging.debug("Running: %s", cmdstr)

        # Real-time output if high debugging level
        if utils.loglevel < logging.DEBUG:
            quiet = False

        p = Popen(cmdstr, shell=True,
            stdin=PIPE if quiet else sys.stdin,
            stdout=PIPE if quiet else sys.stdout, stderr=STDOUT)
        stdoutput = p.communicate()[0]
        stdoutput = stdoutput or ""
        if len(stdoutput) <= MAX_OUTPUT_LOGGING:
            logging.debug("\n" + stdoutput)
        if p.returncode != 0:
            raise ExternalProcessError(self.binary, p.returncode, stdoutput, self.retcodeMap.get(p.returncode, 1))

        return stdoutput

class PackageManagerUtil(CmdLineUtil):

    def __init__(self, rootDir="", adminDir="", binary=""):
        CmdLineUtil.__init__(self, binary)
        # Make sure that paths are compatible with dpkg
        self.rootDir = rootDir.replace("\\", "/")
        self.adminDir = adminDir.replace("\\", "/")

class Dpkg(PackageManagerUtil):

    def __init__(self, rootDir="", adminDir="", options=None):
        PackageManagerUtil.__init__(self, rootDir, adminDir, "dpkg")
        self.retcodeMap = {2: 2}
        self.options = options

    def __call__(self, args, quiet=False):
        if self.options:
            if self.options.force:
                args = "--force-depends --force-overwrite " + args
            else:
                args = "--no-force-downgrade " + args
            args = getDpkgDebugOption(self.options.verbose) + " " + args

        extraDpkgOptions = os.environ.get("BLOCKS_DPKG_OPTIONS")
        if extraDpkgOptions:
            args += " " + extraDpkgOptions

        cmdstr = '--admindir="%s" --instdir="%s" %s' % (self.adminDir, self.rootDir, args)
        return PackageManagerUtil.__call__(self, cmdstr, quiet)

class Apt(PackageManagerUtil):

    def __init__(self, rootDir="", adminDir="", options=None):
        PackageManagerUtil.__init__(self, rootDir, adminDir)
        self.options = options

        # We set APT_CONFIG always in case if BLOCKS_APT_CONFIG variable is not set to
        # the correct apt.conf dir
        aptconf = os.environ.get("BLOCKS_APT_CONFIG")
        if aptconf:
            logging.debug("BLOCKS_APT_CONFIG defined. Using it in place of APT_CONFIG. Value = %s", aptconf)
            os.environ["APT_CONFIG"] = aptconf
        else:
            os.environ["APT_CONFIG"] = os.path.join(self.adminDir, "apt", "apt.conf")

    def __call__(self, args, binary="apt-get", quiet=False):
        self.binary = binary

        dpkgOptions = []
        aptOptions = []

        if binary == "apt-get":
            aptOptions.append("--allow-unauthenticated") # Temporarily activated (when to remove?)

        extraDpkgOptions = ""
        if self.options:
            if self.options.verbose == -1:
                aptOptions.append("-q")
            elif self.options.verbose <= -2:
                aptOptions.append("-qq")

            if self.options.force:
                if binary == "apt-get":
                    aptOptions.append("--force-yes")
                    aptOptions.append("-y")
                    #aptOptions.append("--allow-unauthenticated")
                dpkgOptions.append('-o DPkg::Options::="--force-depends" -o DPkg::Options::="--force-overwrite"')

            extraDpkgOptions = getDpkgDebugOption(self.options.verbose)
            if extraDpkgOptions:
                extraDpkgOptions += " "

        extraDpkgOptions += os.environ.get("BLOCKS_DPKG_OPTIONS") or ""
        if extraDpkgOptions:
            dpkgOptions += ['-o DPkg::Options::="%s"' % o for o in extraDpkgOptions.split()]

        dpkgOptions = " ".join(dpkgOptions)
        aptOptions = " ".join(aptOptions)

        if binary == "apt-get":
            extraAptOptions = os.environ.get("BLOCKS_APTGET_OPTIONS")
        elif binary == "apt-cache":
            extraAptOptions = os.environ.get("BLOCKS_APTCACHE_OPTIONS")
        if extraAptOptions:
            aptOptions += " " + extraAptOptions

        t = Template('-o DPkg::Options::="--instdir=$ROOTDIR" -o DPkg::Options::="--admindir=$ADMINDIR" '
                     '-o DPkg::Options::="--force-bad-verify" '
                     '-o Dir::State="$APTDIR" -o Dir::State::status="$ADMINDIR/status" -o Dir::Cache="$APTDIR/cache" '
                     '-o Dir::Log="$APTDIR" -o Dir::Bin::gzip="$BINDIR/gzip" -o Dir::Bin::dpkg="$BINDIR/dpkg" '
                     '-o Dir::Etc="$APTDIR" -o Dir::Bin::methods="$BINDIR/methods" '
                     '-o Dir::Bin::gpg="$BINDIR/gpgv" '
                     '-o APT::GPGV::TrustedKeyring="$KEYRING" '
                     )
        blocks_gpg = os.environ.get("BLOCKS_GPGTRUSTED")
        if blocks_gpg:
            keyring_path = blocks_gpg
        else:
            keyring_path = os.path.normpath(os.path.join(self.adminDir, "..", "trusted.gpg"))
        cmdstr = t.substitute(ROOTDIR=self.rootDir,
                              ADMINDIR=self.adminDir,
                              APTDIR=self.adminDir + "/apt",
                              BINDIR=utils.getUtilDir(),
                              KEYRING=keyring_path) + dpkgOptions + " " + aptOptions + " " + args

        return PackageManagerUtil.__call__(self, cmdstr, quiet)

def getDpkgDebugOption(level):
    level = min(level, 5)
    return "" if level < 2 else "-D" + "1" * (level - 1)