#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# SourcesList class that manages sources.list
#

from collections import namedtuple
import utils

class RepoError(Exception):
    ''' Repository error '''

Repo = namedtuple("Repo", "name, uri")

class SourcesList(object):
    ''' Manipulate sources.list '''
    def __init__(self, path):
        self.fpath = path
        with open(self.fpath, "a+") as f:
            # parse sources.list to a list of repos
            lines = f.readlines()
            repotuples = zip(lines[::2], lines[1::2])
            self._repos = [Repo(name[1:].strip(), uri.split()[1]) for name, uri in repotuples]

    @property
    def repos(self):
        return self._repos[:]

    def _validateId(self, idnum):
        try:
            idnum = int(idnum)
            if idnum < 1:
                raise ValueError
        except ValueError:
            raise RepoError("Repository id must be positive integer")
        if idnum > len(self._repos):
            raise RepoError("Repository with id %d not found" % idnum)
        return idnum

    def getRepo(self, idnum):
        idnum = self._validateId(idnum)
        return self._repos[idnum - 1]

    def getRepoIdByName(self, name):
        for i, repo in enumerate(self._repos):
            if repo.name == name:
                return i + 1
        raise RepoError("Repository with name '%s' does not exist" % name)

    def add(self, name, uri):
        if name in (repo.name for repo in self._repos):
            raise RepoError("Repository with name '%s' already exists" % name)
        if uri in (repo.uri for repo in self._repos):
            raise RepoError("Repository with URI '%s' already exists" % uri)
        try:
            utils.validateURL(uri)
        except ValueError, ex:
            raise RepoError(str(ex))
        self._repos.append(Repo(name, uri))
        self._write()

    def remove(self, idnum):
        idnum = self._validateId(idnum)
        del self._repos[idnum - 1]
        self._write()

    def removeByName(self, name):
        for i, repo in enumerate(self._repos):
            if repo.name == name:
                del self._repos[i]
                break
        else:
            raise RepoError("Repository with name '%s' does not exist" % name)
        self._write()

    def _write(self):
        with open(self.fpath, "w") as f:
            for name, uri in self._repos:
                f.write("#" + name + "\n")
                f.write("deb " + uri + " /\n")