#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Build data
#

'''
Classes for reading and formatting data from various sources for consumption by
Blocks.Packager and DependencyProcessors.
'''

import os

class BuildData(object):
    '''
    The base class for providing component data to Packager:
        - name of component built
        - location of targetroot
        - location of sourceroot
        - component version
        - target files belonging to the component (relative to targetroot/
        epocroot)
        - source files belonging to the component (relative to sourceroot)
        - file dependencies to other files
    '''

    def __init__(self):
        pass

    def getComponentName(self):
        raise NotImplementedError

    def setComponentName(self):
        raise NotImplementedError

    def getComponentVersion(self):
        raise NotImplementedError

    def setComponentVersion(self):
        raise NotImplementedError

    def getComponentRevision(self):
        raise NotImplementedError

    def setComponentRevision(self, revision):
        raise NotImplementedError

    def getSourceRoot(self):
        raise NotImplementedError

    def setSourceRoot(self):
        raise NotImplementedError

    def getTargetRoot(self):
        raise NotImplementedError

    def setTargetRoot(self):
        raise NotImplementedError

    def getSourceFiles(self):
        raise NotImplementedError

    def addSourceFiles(self):
        raise NotImplementedError

    def getTargetFiles(self):
        raise NotImplementedError

    def addTargetFiles(self):
        raise NotImplementedError

    def setUseEpoch(self):
        raise NotImplementedError

    def hasFiles(self):
        raise NotImplementedError

class BdFile(object):

    def __init__(self, path):
        self.path = os.path.normpath(path)
        if os.sep != "/":
            self.path = "/".join(self.path.split(os.sep))
        self.ownerRequirements = []     # Files that the owner package will
                                        # require
        self.sourceRequirements = []    # Files that the component source
                                        # package will require
        self.variantPlatform = None
        self.variantType = None

    def baseName(self):
        ''' Return the filename without the path. '''
        return os.path.basename(self.path)

    def getPath(self):
        ''' Return path relative to sourceroot/epocroot. '''
        return self.path

    def addSourceDependency(self, path):
        path = os.path.normpath(path)
        if os.sep != "/":
            path = "/".join(path.split(os.sep))
        if path not in self.sourceRequirements:
            self.sourceRequirements.append(path)

    def addOwnerDependency(self, path):
        path = os.path.normpath(path)
        if os.sep != "/":
            path = "/".join(path.split(os.sep))
        if path not in self.ownerRequirements:
            self.ownerRequirements.append(path)

    def setSourceDependencies(self, paths):
        '''
        Set paths of files required by the source package of the component
        owning this Deliverable.

        @type paths: List(String)
        '''
        paths = [os.path.normpath(p) for p in paths]
        if os.sep != "/":
            paths = [ "/".join(path.split(os.sep)) for path in paths ]
        self.sourceRequirements = paths

    def setOwnerDependencies(self, paths):
        '''
        Set paths of files that the package owning this Deliverable depends
        on.

        @type paths: List(String)
        '''
        paths = [os.path.normpath(p) for p in paths]
        if os.sep != "/":
            paths = [ "/".join(path.split(os.sep)) for path in paths ]
        self.ownerRequirements = paths

    def getSourceDependencies(self):
        '''
        Get paths of files required by the source package of the component
        owning this Deliverable.

        @rtype: List(String)
        '''
        return self.sourceRequirements

    def getOwnerDependencies(self):
        '''
        Get paths of files that the package owning this Deliverable depends
        on.

        @rtype: List(String)
        '''
        return self.ownerRequirements

    def getVariantPlatform(self):
        ''' Get stored platform or try to guess from path '''
        if self.variantPlatform:
            return self.variantPlatform
        else:
            # epoc32/release/armv5/udeb/foo
            # -> "armv5"
            parts = os.path.dirname(self.path).split('/')
            if len(parts) > 2:
                return parts[2]
            else:
                return "noarch"

    def getVariantType(self):
        ''' Get stored variantType or try to guess from path '''
        if self.variantType:
            return self.variantType
        else:
            # epoc32/release/armv5/udeb/foo
            # -> "udeb"
            parts = os.path.dirname(self.path).split('/')
            if parts[:2] == ["epoc32", "release"] and len(parts) > 3:
                return parts[3]
            else:
                return None

class PlainBuildData(BuildData):
    '''
    Implementation of BuildData. To add simple file paths, use the
    L{addTargetFiles} and L{addSourceFiles} methods. To incorporate
    dependency information for a target file, create a BdFile instance and then
    add it using the L{addDeliverable} method.

    Similarly L{getTargetFiles} and L{getSourceFiles} return simple lists of
    paths, whereas L{getDependencies} only returns BdFile instances.
    '''

    def __init__(self):
        BuildData.__init__(self)
        self.sourceRoot = ""
        self.targetRoot = ""
        # keys are paths of files belonging to this component
        # values (if any) are Deliverable objects
        self.sourceFiles = {}
        self.targetFiles = {}
        self.apiMapping = {}
        self.componentName = ""
        self.previousNames = []
        self.componentVersion = ""
        self.componentRevision = None
        self.significantAttributes = ("vendor", "license", "group")
        self.attributes = {}
        # extra data for dependency processors
        self.dependencyData = {}
        self.forcePackageAll = False
        self.useEpoch = False

    def getComponentName(self):
        return self.componentName

    def setComponentName(self, name, previousNames=None):
        '''
        Set component name and optionally previous names of the component

        If you add previous names of the component to previousNames
        dependencies are handled correctly in generated bundles by replacing previous bundles.

        @param name: Name of the component
        @type name: String
        @param previousNames: List of previous names of the component
        @type previousNames: List(String)
        '''
        if name:
            self.componentName = name
        else:
            raise ValueError("Component name not defined")
        self.previousNames = previousNames or []
        assert isinstance(self.previousNames, list), "previousNames must be a list"

    def getComponentVersion(self):
        return self.componentVersion

    def setComponentVersion(self, version):
        if version:
            self.componentVersion = version
        else:
            raise ValueError("Component version not defined")

    def getComponentRevision(self):
        return self.componentRevision

    def setComponentRevision(self, revision):
        if revision:
            self.componentRevision = revision
        else:
            raise ValueError("empty revision")

    def getSourceRoot(self):
        return self.sourceRoot

    def setSourceRoot(self, path):
        if os.sep != "/":
            path = "/".join(path.split(os.sep))
        if path:
            self.sourceRoot = os.path.abspath(path)
        else:
            raise ValueError("Source root not defined")

    def getTargetRoot(self):
        return self.targetRoot

    def setTargetRoot(self, path):
        if os.sep != "/":
            path = "/".join(path.split(os.sep))
        if path:
            self.targetRoot = os.path.abspath(path)
        else:
            raise ValueError("Target root not defined")

    def getSourceFiles(self):
        return self.sourceFiles.keys()

    def normalizePath(self, path):
        '''
        Check and clean up path

        @param path: Relative path (to sourceroot or epocroot), Unix or Windows slashes accepted.
        @type path: String
        @return: Path converted to Unix slashes
        '''
        path = os.path.normpath(path)
        path = "/".join(path.split("\\"))
        if not path or path.isspace():
            raise ValueError, "Path not defined or empty"
        if os.path.isabs(path):
            raise ValueError, "Path should not be absolute (%s)" % path
        if "../" in path:
            raise ValueError, "Path must not contain '../' (%s)" % path
        if not os.path.basename(path):
            raise ValueError, "Path must not contain a filename (%s)" % path
        if path in ("."):
            raise ValueError, "Path must not be '.' (%s)" % path
        return path

    def addSourceFiles(self, list):
        '''
        Add a simple list of filenames without dependency data.

        @param list: List of relative paths to source root.
        @type list: List(String)
        '''
        for path in list:
            path = self.normalizePath(path)
            if path not in self.sourceFiles:
                self.sourceFiles[path] = None

    def getTargetFiles(self):
        return self.targetFiles.keys()

    def addTargetFiles(self, list):
        '''
        Add a simple list of filenames without dependency data.

        @param list: paths relative to epocroot/targetroot
        @type list: List(String)
        '''
        for path in list:
            path = self.normalizePath(path)
            if path not in self.targetFiles:
                self.targetFiles[path] = None

    def addNonstandardInterfaces(self, mapping):
        '''
        Add nonstandard interfaces for binaries

        @param mapping: Update mappings of binaries to correct api files (.dso).
        Can be either dictionary or iterable of key/value pairs.
        Note: Can be called multiple times to add mappings
        @type mapping: Dict(string)
        '''
        self.apiMapping.update(mapping)

    def getDependencies(self):
        return [file for file in self.targetFiles.values() if file is not None]

    def addDeliverable(self, deliverable):
        '''
        Add Deliverable that also contains dependency data.

        @param deliverable: A BdFile instance containing the file path and
        additional metadata.
        @type deliverable: BdFile
        '''
        deliverable.path = self.normalizePath(deliverable.getPath())
        self.targetFiles[deliverable.path] = deliverable

    def setUseEpoch(self):
        self.useEpoch = True

    def getUseEpoch(self):
        return self.useEpoch

    def hasFiles(self):
        return bool(self.targetFiles) or bool(self.sourceFiles)