#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Logging system
#

''' Logging '''

import logging
import os
import multiprocessing

class NullHandler(logging.Handler):
    def __init__(self, level=None):
        logging.Handler.__init__(self, level=level)

    def emit(self, value):
        pass

    def setLevel(self, level):
        pass

class Logging(object):

    STATUSNAME = "pfw-status"
    STATUSLEVEL = logging.DEBUG

    logLevel = None
    logFormat = '%(asctime)s %(levelname)s %(name)s %(message)s'
    logTimeFormat = '%Y-%m-%d %H:%M:%S'
    logDir = None
    doStatusLogs = False
    defaultHandler = None
    lock = None

    @classmethod
    def getLogger(cls, name="pfw", level=None, addHandler=True):
        '''
        Get a pre-configured logger with handler etc.
        '''
        l = logging.getLogger(name)
        l.propagate = False
        l.setLevel(level or cls.logLevel)

        if addHandler:
            if not l.handlers:
                l.addHandler(cls.getDefaultHandler())
        return l

    @classmethod
    def getStatusLogger(cls):
        l = cls.getLogger(name=cls.STATUSNAME, level=cls.STATUSLEVEL, addHandler=False)
        statusHandler = logging.StreamHandler()
        statusHandler.setFormatter(logging.Formatter('Status: %(message)s', cls.logTimeFormat))
        if not l.handlers:
            l.addHandler(statusHandler)
        if not cls.doStatusLogs:
            statusHandler.addFilter(logging.Filter("nevergonnamatch"))
        return l

    @classmethod
    def getDefaultHandler(cls):
        if cls.logLevel == None:
            # prevent "no handler" messages
            defaultHandler = NullHandler()
        elif cls.logDir:
            defaultHandler = logging.FileHandler(os.path.join(cls.logDir, "pfw.log"))
        elif cls.defaultHandler:
            defaultHandler = cls.defaultHandler
        else:
            defaultHandler = logging.StreamHandler()
        if cls.logFormat and defaultHandler.formatter is None:
            defaultHandler.setFormatter(logging.Formatter(cls.logFormat, cls.logTimeFormat))

        # Wrap handlers handle method with lock if lock is given
        if cls.lock:
            def lockWrapper(method):
                def lockCall(*args, **kwargs):
                    with cls.lock:
                        ret = method(*args, **kwargs)
                    return ret
                return lockCall
            defaultHandler.handle = lockWrapper(defaultHandler.handle)

        return defaultHandler

    @classmethod
    def setupLogHandling(cls, logLevel=None, logDir=None, logFormat=None, logTimeFormat=None,
        doStatusLogs=None, handler=None, lock=None):

        if logLevel != None:
            cls.logLevel = logLevel

        if logDir != None:
            cls.logDir = logDir

        if logFormat != None:
            cls.logFormat = logFormat

        if logTimeFormat != None:
            cls.logTimeFormat = logTimeFormat

        if doStatusLogs != None:
            cls.doStatusLogs = doStatusLogs

        if handler != None:
            cls.defaultHandler = handler

        if lock is None:
            cls.lock = multiprocessing.Lock()
        else:
            cls.lock = lock