#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Package model
#

'''
PackageModel contains a hierarchy of classes that store metadata about files
for use by PackageWriters::

    + Component [ Files that together provide some functionality are grouped as
    |             a Component. ]
    |
    -- + Package [ The files in a component are sorted together by type, e.g.
       |           sources, executables or libraries, into Packages. A Package
       |           is stored as one archive, e.g. armv5 executables in component
       |           foo could be written into foo.exec-arm-VERSION.deb ]
       |
       -- Deliverable [ A file ]

A L{Packager} uses L{Blocks.Packaging.Rules} to determine which subclass of Deliverable to use
when storing file metadata, as well as which Package it belongs to.
'''

from cPickle import dump, load
import os

try:
    from hashlib import md5
except ImportError:
    import md5

from Blocks.Packaging.PackageWriter import Packageable, FileSet

class Component(Packageable):
    ''' A Component is used to store Packages and associated metadata. '''

    def __init__(self, name, version, packages=None):
        '''
        @param name: Component name
        @param version: Component version
        @param packages: list of packages belonging to component
        @type name: String
        @type version: Version string
        @type packages: List(Package)
        '''
        Packageable.__init__(self)
        self.name = name
        if version:
            self.setVersion(version)
        self.__attributes = {}
        self.__packages = {}
        packages = packages or []
        for p in packages:
            self.addPackage(p)

    def __eq__(self, other):
        if self.getName() == other.getName() \
            and self.getSignificantAttributes() == other.getSignificantAttributes() \
            and self.getPackages() == other.getPackages():
            return True
        else:
            return False

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        out = "Component:\n\
        name: %s\n\
        version: %s\n\
        buildId: %s\n\
        filecount: %s\n\
        packageAllFlag: %s\n\
        attributes: %s\n\
        packages: %s\n" % (
            self.name,
            self.getVersion(),
            self.getBuild(),
            len(self.getFiles()),
            self.getPackageAllFlag(),
            [str(k) + ":" + str(v) for k, v in self.getComponentAttributes().items()],
            [p.getName() for p in self.getPackages()])
        return out

    @staticmethod
    def load(f):
        '''
        Read a previously pickled component from a file.

        @param f: file or file-like object
        @return: A Component loaded from f
        @rtype: Component
        '''
        if isinstance(f, basestring) and os.path.isfile(f):
            f = open(f, "r")
        return load(f)

    def dump(self, f):
        '''
        Dump the Component using pickle

        @param f: filename or file-like object
        '''
        if isinstance(f, basestring):
            f = open(f, "wb")
        dump(self, f, 0)

    def getFiles(self):
        '''
        Return list of files belonging to the member packages.

        @return: files
        @rtype: List(Deliverable)
        '''
        allFiles = []
        for p in self.getPackages():
            allFiles += p.getFiles()
        allFiles.sort()
        return allFiles

    def addPackage(self, package):
        '''
        @param package:
        @type package: Package
        '''
        if type(package) != Package:
            raise ValueError("Not a Package: %s" % type(package))
        elif package.getName() in self.__packages:
            raise ValueError("A package with that name exists already")
        else:
            self.__packages[package.getName()] = package

    def getPackage(self, name):
        '''Get package by name.

        @param name: The package name
        @type name: String
        @return: Package
        @rtype: Package
        '''
        if name == "":
            raise ValueError("getPackage: name cannot be an empty string")
        return self.__packages.get(name)

    def getPackages(self):
        '''
        Return list of all packages.

        @return: The packages that make up the component
        @rtype: List(Package)
        '''
        p = self.__packages.values()
        p.sort(lambda x, y: cmp(x.getName(), y.getName()))
        return p

    def getBinaryPackage(self):
        for package in self.getPackages():
            if package.getName().count(".exec-"):
                return package
        return None

    def getTracePackage(self):
        for package in self.getPackages():
            if package.getName().count("trace"):
                return package
        return None

    def getName(self):
        '''
        @return: name
        @rtype: String
        '''
        return self.name

    def getComponentAttributes(self):
        '''
        Get attributes in the raw, as ComponentAttributes.

        @return: Attributes
        @rtype: Dictionary(ComponentAttribute)
        '''
        return self.__attributes

    def getSignificantAttributes(self):
        '''
        Get attributes that are significant for packaging purposes.

        @return: Attributes
        @rtype: Dictionary(any)
        '''
        ret = {}
        for key in self.__attributes.keys():
            if self.__attributes[key].significant:
                ret[key] = self.__attributes[key].value
        return ret

    def getAttributes(self):
        '''
        Get attributes stripped of ComponentAttribute metadata.

        @return: Attributes
        @rtype: Dictionary(Any)
        '''
        ret = {}
        for key in self.__attributes.keys():
            ret[key] = self.__attributes[key].value
        return ret

    def setAttribute(self, name, value, significant=False):
        '''
        Any type can be passed as a value, but attributes are stored internally
        as ComponentAttributes. When setting attributes that are significant for
        packaging rules, i.e. the value is taken into consideration when
        comparing the equality of two Components, you must wrap the value in a
        ComponentAttribute or specify significant as True.

        @param name: Key for attribute
        @param value: Value for attribute
        '''
        if not isinstance(value, ComponentAttribute):
            value = ComponentAttribute(value, significant)
        self.__attributes[name] = value

    def diff(self, other):
        '''
        Get details of differences between packages. For quick comparison just
        use the normal comparison operators == and !=.

        @param other: Another Component
        @return: less, more, same, different;
         - Names of Packages not present in self
         - Names of Packages not present in other
         - Names of Packages that are the same in both
         - Names of Packages that are present in both but differ
        @rtype: (List, List, List, List)
        '''

        if not isinstance(other, Component):
            raise ValueError, "Other is not a Component"

        less = []   # What packages does other have that I don't?
        more = []   # What packages do I have that other doesn't?
        same = []
        different = []

        for o in other.getPackages():
            p = self.getPackage(o.getName())
            if not p:
                less.append(o.getName())
            elif o == p:
                same.append(o.getName())
            else:
                different.append(o.getName())

        for p in self.getPackages():
            if p.getName() not in less + same + different:
                more.append(p.getName())

        return (less, more, same, different)

class ComponentAttribute(object):
    '''
    Store extra metadata about Component attributes in addition to the attribute
    value. The main motivation for using ComponentAttribute is to include a flag
    indicating whether the attribute is relevant when comparing the equality of
    two Components.
    '''
    def __init__(self, value, significant=False):
        if type(significant) != bool:
            raise TypeError("ComponentAttribute.significant must be a Boolean")
        # you've got to be kidding
        if isinstance(value, ComponentAttribute):
            raise TypeError("Trying to play nested dolls with ComponentAttribute.")
        self.value = value
        self.significant = significant

    def __eq__(self, other):
        return isinstance(other, ComponentAttribute) and self.value == other.value

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        if self.significant:
            return str(self.value)+"(S)"
        else:
            return str(self.value)

class Package(FileSet):
    '''
    A Package is essentially a named list of Deliverables with associated
    metadata.
    '''
    def __init__(self, name, files=None):
        '''

        @param name: Package name
        @type name: String
        @param files: List of Deliverables
        @type files: List(Deliverable)
        '''
        FileSet.__init__(self)
        if not name:
            raise ValueError('Unacceptable name "%s" for a Package' % name)
        self.name = name
        self.files = {}
        for f in files or []:
            self.addFile(f)
        self.type = None
        self.arch = None
        self.group = None
        self.depends = []       # Dependencies to list the Packages this one depends on
        self.preconditions = [] # This is packaged only if any of these packages are flagged for packaging
        self.replaces = []

    def __eq__(self, other):
        return self.name == other.name and self.getFilesSorted() == other.getFilesSorted()

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return "Package:\n\
        name: %s\n\
        filecount: %s\n\
        type: %s\n\
        versioninfo: %s\n\
        arch: %s\n\
        group: %s\n\
        packageFlag: %s\n\
        lastPackagedVersion: %s\n\
        apiVersion: %s\n\
        depends: %s\n\
        preconditions: %s\n" % (
            self.name,
            len(self.getFiles()),
            self.type,
            self.getIdentifier(),
            self.arch,
            self.group,
            self.getPackageFlag(),
            self.getLastPackagedVersion(),
            self.getApiVersion(),
            [str(v) for v in self.getDependencies()],
            self.preconditions)

    def diff(self, other):
        '''
        Get details of differences between packages. For quick comparison just
        use the normal comparison operators == and !=.

        @param other: Another Package
        @return: less, more, same, different;
         - Paths of Deliverables not present in self
         - Paths of Deliverables not present in other
         - Paths of Deliverables of Deliverables that are the same in both
         - Paths of Deliverables that are present in both but differ
        @rtype: (List, List, List, List)
        '''

        if not isinstance(other, Package):
            raise ValueError("Other is not a Package")

        less = []   # What does other Package have that I don't?
        more = []   # What do I have that other doesn't?
        same = []
        different = []

        for o in other.getFiles():
            f = self.getFile(o.path)
            if not f:
                less.append(o.path)
            elif o == f:
                same.append(o.path)
            else:
                different.append(o.path)

        for f in self.getFiles():
            if f.path not in less + same + different:
                more.append(f.path)

        return (less, more, same, different)

    def getFiles(self):
        '''
        Return list of files belonging to this package.

        @return: files
        @rtype: List(Deliverable)
        '''
        return self.files.values()

    def getFilesSorted(self):
        '''
        Return list of files belonging to this package, sorted by path

        @return: files
        @rtype: List(Deliverable)
        '''
        keys = self.files.keys()
        keys.sort()
        return [self.getFile(k) for k in keys]

    def getFile(self, path):
        '''
        Get a file by path
        @param path: path
        '''
        return self.files.get(path)

    def addFile(self, file):
        '''
        Add a new Deliverable to the package

        @param file: The file to be added
        @type file: Deliverable
        @return: True on success, False if a Deliverable with the path already exists
        @raise TypeError: Tried to add a non-Deliverable
        '''
        if isinstance(file, Deliverable):
            if not self.getFile(file.path):
                self.files[file.path] = file
                return True
            else:
                return False
        else:
            raise TypeError("Will only add Deliverables to package. (%s)" % type(file))

    def getName(self):
        '''
        @return: name
        @rtype: String
        '''
        return self.name

    def getDependencies(self):
        '''
        @return: dependencies
        @rtype: List(Dependency)
        '''
        return self.depends

    def addDependency(self, dependency):
        if dependency not in self.depends and dependency.package != self.getName():
            self.depends.append(dependency)
            return True
        else:
            return False

    def getApiFingerprint(self):
        '''
        @return: API Fingerprint, a 12-character unique identifer for the API
        @rtype: String
        '''
        apis = {}
        filenames = []
        apiKeys = []
        separator = "\t"
        if hasattr(md5, "new"):
            m = md5.new()
        else:
            m = md5()
        if self.getApiVersion():
            for f in self.getFiles():
                if (isinstance(f, Executable) and f.api):
                    # The assumption is that executables with the same name have
                    # the same API, so only include one of them.
                    apis[os.path.basename(f.path)] = f.api.interface
            filenames = apis.keys()
            filenames.sort()
            for filename in filenames:
                m.update(filename + separator)
                apiKeys = apis[filename].keys()
                apiKeys.sort()
                for apiKey in apiKeys:
                    m.update(str(apiKey) + str(apis[filename][apiKey]))
                m.update(separator)
            return m.hexdigest()[:12]
        else:
            return None

class Deliverable(object):
    ''' A class for storing metadata about an individual file. '''
    def __init__(self, path, type="file", size=None, checksum=None):
        '''
        The type can be omitted, by default it is "file"

        @param path: The relative location of the file
        @param type: The type of Deliverable
        @param size: Size in bytes
        @param checksum: md5 hexdigest
        @type path: String
        @type type: String
        @type size: Integer
        @type checksum: String
        '''
        self.path = path
        self.type = type
        self.size = size
        self.checksum = checksum

    def __eq__(self, other):
        if isinstance(other, Deliverable):
            return self.path == other.path and self.checksum == other.checksum

    def __ne__(self, other):
        return not self == other

    # TODO: Are these extra comparison operators necessary?
    def __lt__(self, other):
        if self.path < other.path:
            return True
        elif self.path == other.path:
            if self.checksum < other.checksum:
                return True
        return False

    def __gt__(self, other):
        if self.path > other.path:
            return True
        elif self.path == other.path:
            if self.checksum > other.checksum:
                return True
        return False

    def __str__(self):
        return "Deliverable:\npath: %s\ntype: %s\nsize: %s\nchecksum: %s\n" % (self.path, self.type, self.size, self.checksum)

    def getPath(self):
        return self.path

    def getChecksum(self):
        return self.checksum

    def baseName(self):
        return os.path.basename(self.path)

class PEBinary(Deliverable):
    ''' A Deliverable with additional arch, mappath and sympath attributes. '''

    def __init__(self, path, type, size=None, checksum=None):
        '''
        The type is mandatory

        @param path: The relative location of the file
        @param type: The type of Deliverable
        @type path: String
        @type type: String
        '''
        Deliverable.__init__(self, path, type, size, checksum)
        self.arch = None
        self.mappath = None

    def __str__(self):
        out = Deliverable.__str__(self).replace("Deliverable:", "PEBinary:")
        out += """arch: %s\nmappath: %s\n""" % (self.arch, self.mappath)
        return out

class E32Binary(Deliverable):
    ''' A Deliverable with additional arch, mappath and sympath attributes. '''

    def __init__(self, path, type, size=None, checksum=None, variant=None):
        '''
        The type is mandatory

        @param path: The relative location of the file
        @param type: The type of Deliverable
        @type path: String
        @type type: String
        '''
        Deliverable.__init__(self, path, type, size, checksum)
        self.variant = variant
        self.arch = None
        self.mappath = None
        self.sympath = None

    def __str__(self):
        out = Deliverable.__str__(self).replace("Deliverable:", "E32Binary:")
        out += "variant: %s\narch: %s\nmappath: %s\nsympath: %s\n" % (self.variant, self.arch, self.mappath, self.sympath)
        return out

class Resource(Deliverable):
    ''' A Deliverable with target, targetpath and source attributes. '''
    def __init__(self, path, type, size=None, checksum=None):
        '''
        @param path: The relative location of the file
        @param type: The type of Deliverable
        @type path: String
        @type type: String
        '''
        Deliverable.__init__(self, path, type, size, checksum)
        self.target = None
        self.targetpath = None
        self.source = None

    def __str__(self):
        out = Deliverable.__str__(self).replace("Deliverable:","Resource:")
        out += "target: %s\ntargetpath: %s\nsource: %s\n" % (self.target, self.targetpath, self.source)
        return out

class Executable(E32Binary):
    ''' An E32Binary with an optional API attribute. The API is obtained from the associated DSO file.

    Examples: exe, plugin or dll
    '''
    def __init__(self, path, type, size=None, checksum=None, variant=None, api=None):
        '''
        @param path: The relative location of the file
        @param type: The type of Executable
        @type path: String
        @type type: String
        '''
        E32Binary.__init__(self, path, type, size, checksum, variant)
        if api:
            if not isinstance(api, API):
                raise ValueError("The api must be an instance of API")
        self.api = api

    def __str__(self):
        out = E32Binary.__str__(self).replace("E32Binary:","Executable:")
        out += "api: %s\n" % self.api
        return out

class Library(E32Binary):
    ''' A non-executable binary.

    Examples: staticlib or dso
    '''
    def __init__(self, path, type, size=None, checksum=None, variant=None):
        '''
        @param path: The relative location of the file
        @param type: The type of Library
        @type path: String
        @type type: String
        '''
        E32Binary.__init__(self, path, type, size, checksum, variant)

    def variantPlatform(self):
        return self.path.lstrip(os.sep).split(os.sep)[2]

    def variantType(self):
        out = self.path.lstrip(os.sep).split(os.sep)[3]
        if out == "lib":
            return None
        return out

class API(object):
    ''' An API defines an interface and a version, with methods for comparison to another API. '''
    def __init__(self, version="1.0", interface=None):
        if type(version) != str:
            raise TypeError("The version for an API must be a string")
        if not version:
            raise ValueError("API version cannot be an empty string")
        if type(interface) != dict:
            raise TypeError("The interface for an API must be a dictionary")
        for k, v in interface.items():
            if not (k and v):
                raise ValueError("Cannot have empty keys or values in an interface")
        self.version = version
        self.interface = interface

    def __str__(self):
        return "API:{version: %s, interface_keycount: %s}" % (self.version, len(self.interface.keys()))

    def provides(self, other):
        ''' This API is compatible with other. This interface may be identical to or a superset of other.

        @param other: Another API to compare with
        @type other: API
        @return: This provides same functions as other
        @rtype: Boolean
        '''
        ret = True
        for index, function in other.interface.items():
            if index in self.interface:
                if self.interface[index] != function:
                    ret = False
            else:
                ret = False
        return ret

    def isIdentical(self, other):
        ''' The interface exports the exact same functions as other.

        @param other: Another API to compare with
        @type other: API
        @return: This API is identical to other
        @rtype: Boolean
        '''
        return self.interface == other.interface
